require 'rubygems'
require 'hpricot'

module Scrubyt
  ##
  #=<tt>Group more filters into one</tt>
  #
  #Server as an umbrella for filters which are conceptually extracting 
  #the same thing - for example a price or a title or ...
  #
  #Sometimes the same piece of information can not be extracted with one filter 
  #across more result instances (for example a price has an XPath in record n,
  #but since in record n+1 has a discount price as well, the real price is pushed
  #to a different XPath etc) - in this case the more filters which extract the same
  #thing are hold in the same pattern.
  class Pattern
    #Type of the pattern; 

    # a root pattern represents a (surprise!) root pattern
    PATTERN_TYPE_ROOT = 0
    # a tree pattern represents a HTML region
    PATTERN_TYPE_TREE = 1
    # represents an attribute of the node extracted by the parent pattern
    PATTERN_TYPE_ATTRIBUTE = 2
    # represents a pattern which filters its output with a regexp
    PATTERN_TYPE_REGEXP = 3    

    #The pattern can be either a model pattern (in this case it is
    #written to the output) or a temp pattern (in this case it is skipped)
    #Will be implemented in a higher version (i.e. not 0.1.0) - for now, everything
    #is considered to be a model pattern
    
    #Model pattern are shown in the output
    OUTPUT_TYPE_MODEL = 0
    #Temp patterns are skipped in the output (their ancestors are appended to the parent
    #of the pattrern which was skipped
    OUTPUT_TYPE_TEMP = 1
    
    #These fields can be set upon wrapper creation - i.e. a field which is public but not contained here can be accessed 
    #from outside, but not set as a result of wrapper construction
    SETTABLE_FIELDS = ['generalize', 'type', 'output_type', 'example']
    
    attr_accessor :name, :output_type, :generalize, :children, :filters, :parent, 
                  :last_result, :result, :example, :limit, 
                  :examples, :parent_of_leaf, :evaluation_context,
                  :indices_to_extract, :evaluation_context
    attr_reader :type, :generalize_set, :next_page_url, :result_indexer

    def initialize (name, *args)
      @name = name                #name of the pattern
      parse_args(args)            #parse the rest of the arguments
      @children = []              #child patterns
      @filters = []               #filters of the wrapper
      @result = Result.new        #hierarchical results of the pattern
      if @examples == nil
        filters << Scrubyt::Filter.new(self) #create a default filter
      else
        @examples.each do |example|
          filters << Scrubyt::Filter.new(self,example) #create a filter
        end
      end      
    end
    
    #Parse the args passed as *args; There is only one compulsory parameter to pattern: it's name
    #All the other parameters can (but do not have to) be specified;
    #
    #If an example is specified, it *MUST* be the first parameter; the order of the other
    #parameters is irrelevant
    def parse_args(args)
      #Grab any examples that are defined!
      look_for_examples(args)
      args.each do |arg|
        next if !arg.is_a? Hash
        arg.each do |k,v|
          #Set only the setable fields
          if SETTABLE_FIELDS.include? k.to_s 
            #If the user is specifying a pattern type, turn it into the corresponding constant
            v = "PATTERN_TYPE_#{v.to_s.upcase!}" if k.to_s == 'type'
            v = "OUTPUT_TYPE_#{v.to_s.upcase!}" if k.to_s == 'output_type'
            #Otherwise, if nothing special is happening, isntance_eval the hash pair
            instance_eval("@#{k.to_s} = #{v}") 
          end
          #This flags says that the user explicitly wants to set generalization on a pattern
          #In this case, of course, our heuristics do not apply - the users setting overrides
          #it
          @generalize_set = true if (k.to_s == 'generalize')        
        end
      end
      #default settings - the user can override them, but if she did not do so,
      #we will setup some meaningful defaults
      @type ||= PATTERN_TYPE_TREE
      @output_type ||= OUTPUT_TYPE_MODEL
      #don't generalize by default
      @generalize ||= false
      #This flag indicates that the user set 'generalize' to some value;
      #This way we can ensure that the explicit setting will not be overridden 
      @generalize_set ||= false 
    end    
    
    #Dispatcher function; The class was already too big so I have decided to factor
    #out some methods based on their functionality (like output, adding constraints) 
    #to utility classes.
    #
    #The second function besides dispatching is to lookup the results in an evaluated 
    #wrapper, for example
    #
    # camera_data.item[1].item_name[0]
    def method_missing(method_name, *args, &block)
      case method_name.to_s
      when 'select_indices'
        @result_indexer = Scrubyt::ResultIndexer.new(*args)
        self
      when /^to_/
        Scrubyt::ResultDumper.send(method_name.to_s, self)
      when /^ensure_/
        Scrubyt::ConstraintAdder.send(method_name, self, *args)
      else
        @children.each { |child| return child if child.name == method_name.to_s }
        nil
      end
    end

    #Companion function to the previous one (Pattern::method_missing). It makes
    #inspecting results, like
    #
    #    camera_data.item[1].item_name[0]
    #    
    #possible. The method Pattern::method missing handles the 'item', 'item_name' etc.
    #parts, while the indexing ([1], [0]) is handled by this function.
    #If you would like to select a different document than the first one (which is
    #the default), you should use the form:
    #
    #    camera_data[1].item[1].item_name[0]
    def [](index)
      if @name == 'root'
        @evaluation_context.document_index = index
      else
        @parent.last_result = @parent.last_result[@evaluation_context.document_index] if @parent.last_result.is_a? Array
        return nil if (@result.lookup(@parent.last_result)) == nil
        @last_result = @result.lookup(@parent.last_result)[index]
      end
      self    
    end
    
    ##
    #If export is called on the root pattern, it exports the whole extractor wher it is
    #defined; See export.rb for further details on the parameters
    def export(file, output_file_name=nil, extractor_result_file_name=nil)
      Scrubyt::Export.export(file, self, output_file_name, extractor_result_file_name)
    end
    
    ##
    #Add a filter to this pattern    
    def add_filter(filter)
      @filters << filter
      return self
    end

    ##
    #Add a child pattern to this pattern
    def add_child_pattern(child)
      child.parent = self
      #by default, generalize direct children of the root pattern, but only in the case if
      #@generalize was not set up explicitly
      child.generalize = true if (!child.generalize_set && child.parent != nil && child.parent.parent == nil)
      @children << child
    end
        
    ##
    #Evaluate the pattern. This means evaluating all the filters and adding 
    #their extracted instances to the array of results of this pattern
    def evaluate      
      #No need to evaluate if there is no parent pattern
      return if @parent == nil
      all_filter_results = []
      @filters.each do |filter|
        filter_index = @filters.index(filter)
        filter_index = 0 if @parent.filters.size <= filter_index
        filter.source = @parent.filters[filter_index].sink
        filter.source.each do |source|          
          r = filter.evaluate(source)
          next if r == nil
          if filter.constraints.size > 0
            #in the beginning, keys of result_hash are made up of all the results of the filter
            #with value = true; Later on, only those results will have 'true' value which are
            #accepted with all filters
            result_hash = {}
            r.each { |res| result_hash[res] = true }
            result_hash.keys.each do |res|
              filter.constraints.each { |constraint| result_hash[res] &&= constraint.check(res) }
            end
            result = result_hash.reject {|k,v| k if !v}
            sorted_result = r.reject {|e| !result.keys.include? e}
            indexer = @result_indexer == nil ? sorted_result : @result_indexer.select_indices_to_extract(sorted_result)
            add_result(filter, source, indexer)
          else
            indexer = @result_indexer == nil ? r : @result_indexer.select_indices_to_extract(r)
            add_result(filter, source, indexer)
          end#end of constraint check
        end#end of source iteration
      end#end of filter iteration    
    end   
    
    def get_constraints
      filters[0].constraints
    end     
       
private        
    def look_for_examples(args)
      if (args[0].is_a? String)
        @examples = args.select {|e| e.is_a? String}
        #Check if all the String parameters are really the first
        #parameters 
        args[0..@examples.size-1].each do |example|
          if !example.is_a? String
            puts 'FATAL: Problem with example specification'
          end
        end
      elsif (args[0].is_a? Regexp)
        @examples = args.select {|e| e.is_a? Regexp}
        #Check if all the String parameters are really the first
        #parameters 
        args[0..@examples.size].each do |example|
          if !example.is_a? Regexp
            puts 'FATAL: Problem with example specification'
          end
        end
        @type = PATTERN_TYPE_REGEXP
      end
    end
    
    def add_result(filter, source, results)
      results.each do |res|
        filter.sink << res
        @result.add_result(source, res)
      end      
    end # end of method generate_examples
  end #end of class Pattern
end #end of module Scrubyt
