#require File.join(File.dirname(__FILE__), 'pattern.rb')
 
module Scrubyt
  # =<tt>exporting previously defined extractors</tt>
  class Export
    ##
    #Exports the given extractor (specified by it's root pattern) from the given file
    #
    #_input_file_ - the full path of the file where the extractor was defined. This can
    #be achieved by calling 
    # 
    #  pattern.export(__File__)
    # 
    #from the file of the extractor definition.
    # 
    #*parameters*
    # 
    #_pattern_ - the root pattern of the extractor. This is the variable 'something' in
    #such a call:
    # 
    #  something = Scrubyt::Extractor.define ...
    #  
    #However, since the export method should not be called directly (pattern is calling 
    #it), you will probably never need to care about this parameter.
    #
    #_output_file_name_ - the name of the file where the exported extractor should be 
    #dumped; From default (i.e. if you don't specify this parameter) this is 
    #"#{wrapper_name}_extractor_export.rb". You may override this setting if specifying
    #this optional parameter.
    #
    #_extractor_result_file_name_ - the name of the file, where the result of the 
    #*exported* extractor should be dumped - for example, if _output_file_name_ is "foo.rb"
    #and _extractor_result_file_name_ is "bar.xml", the extractor is exported to a file named
    #"foo.rb", and after running "foo.rb", the results will be dumped to the file "bar.xml"
    #If this option is not specified, the result is dumped to standard output as XML.
    #
    #Examples:
    #
    #  camera_data = Scrubyt::Extractor.define do
    #    Action.fetch File.join(File.dirname(__FILE__), "input.html")
    #    
    #    P.item_name "Canon EOS 20D SLR Digital Camera (Lens sold separately)"
    #  end
    #  
    #  camera_data.export(__FILE__)
    #  
    #This will export this extractor to a file called "camera_data_extractor_export.rb". 
    #If "camera_data_extractor_export.rb" will be executed, the result will be dumped
    #to the standard output.
    #  
    #Note that the export method in the last line belongs to the class Scrubyt::Pattern
    #and not to Scrubyt::Export (i.e. this class). Scrubyt::Pattern.export will call 
    #Scrubyt::Export.export.
    #
    #  camera_data = Scrubyt::Extractor.define do
    #    Action.fetch File.join(File.dirname(__FILE__), "input.html")
    #    
    #    P.item_name "Canon EOS 20D SLR Digital Camera (Lens sold separately)"
    #  end
    #  
    #  camera_data.export(__FILE__, 'my_super_camera_extractor.rb', '/home/peter/stuff/result.xml')
    #  
    #This snippet will export the extractor to a file named 'my_super_camera_extractor.rb'.
    #After running 'my_super_camera_extractor.rb', the result will be dumped to the file
    #'/home/peter/stuff/result.xml'.
    def self.export(input_file, pattern, output_file_name, extractor_result_file_name)
      @result = ""
      contents = open(input_file).read
      wrapper_name = contents.scan(/\s+(.+)\s+=.*Extractor\.define.*/)
      output_file = output_file_name == nil ? open("#{wrapper_name}_extractor_export.rb", 'w') :
                                              open(output_file_name, 'w')
      export_header(output_file)
      export_extractor(contents, pattern, output_file)
      export_footer(output_file, wrapper_name, extractor_result_file_name)
      cleanup_result
      output_file.write(@result)      
      output_file.close
      @result    
    end

private
    def self.export_header(output_file)
      @result += "require 'rubygems'\nrequire 'scrubyt'\n\n"
    end
    
    def self.cleanup_result
      @result.gsub!('P.') {}
    end
   
    #OK, I have to admit: this function is powered by woodo magic. A lots of woodoo magic. 
    #Piles of tons of heaps of woodoo magic :-)
    #
    #The only reason I can expect it to work is that it passes all the tests of the extractors 
    #I have created so far. However at the same time  I know how to create one easily which 
    #would break the exporting, so don't experiment with this too much...
    #
    #The other solutions include:
    #- serialization (yaml, pstore etc) but that would mess the code terribly up - so 
    #therefore I did not chose this solution.
    #- defining the block as string - however, this introduces ugly %q{}s etc - all in all, 
    #this is still a more viable solution that serialization IMHO
    #- a lot of other tricks - however, all of these introduce a lot of noise which I don't
    #like.
    #
    #Conclusion: If there will be no terrible, unrepairable, uncontrollable etc. problems 
    #with this approach, it will be replaced (probably with constructing the extractor as
    #a string). However, until that point, it will stay.    
    def self.export_extractor(contents, pattern, output_file)
      first_line = contents.scan(/.*Extractor\.define.*/)
      #During wrapper construction, we count the number of blocks; add one occurrence of 
      #end (to close the block of the extractor definition)
      count = pattern.evaluation_context.block_count + 1
      #Construct the extractor definition matching regexp based on the number of ends
      definition = contents.scan(/Extractor\.define(?:.*?(?:\}|end)){#{count.to_s}}/m)
      #Since the regexp matching the extractor definition was multiline, get the first
      #line separately and patch it in!
      rows = definition[0].split("\n")
      #Lord of the hacks :-) Originally, when we have used the class P as a pattern definer,
      #patterns could be matched very easily from the extractor definition (because they begun
      #with 'P.'). Now that P has been removed, mimick it!       
      rows.each do |row|       
        #Do not prepend P. to comments and empty lines
        next if (row.strip =~ /^#/ || row.strip == '')
        #Do not prepend P. to any of the reserved keywords 
        jump_to_next = false
        NavigationActions::KEYWORDS.each { |keyword| jump_to_next = true if row.strip =~ /^#{keyword}/ }
        next if jump_to_next
        #Prepend P.s - the hairy stuff with eval et al is there to preserve the whitespace
        row.sub!(/\s+(.*)/) { "#{' ' * eval((row.size - row.lstrip.size ).to_s)} P.#{$1}" } 
        #Don't forget also the stuff in parentheses!
        row.gsub!(/\{\s+/) {"{P."}
      end
      rows[0] = first_line
      #@full_definition holds the original definition (at this point, later on it will be
      #gsub!bed and all)
      @full_definition = rows.join("\n")
      #This hash contains all the examples that need to be replaced with their XPath 
      #counterparts;"P.#{name}"
      #We are relying on the convention that if an example is definied, it is always
      #the first parameter and it is always a string 
      @name_to_xpath_map = {}
      create_name_to_xpath_map(pattern)
      #Replace the examples which are quoted with " and '
      @name_to_xpath_map.each do |name, xpaths|
        replace_example_with_xpath(name, xpaths, %q{"})
        replace_example_with_xpath(name, xpaths, %q{'})
      end
      #Finally, add XPaths to pattern which had no example at the beginning (the XPath was
      #generated from the child patterns
      @name_to_xpath_map.each do |name, xpaths| 
        xpaths.reverse.each do |xpath|
          comma = @full_definition.scan(Regexp.new("P.#{name}(.+)$"))[0][0].sub('do'){}.strip == '' ? '' : ','
          if (@full_definition.scan(Regexp.new("P.#{name}(.+)$"))[0][0]).include?('{')
            @full_definition.sub!("P.#{name}") {"P.#{name}('#{xpath}')"}
          else
            @full_definition.sub!("P.#{name}") {"P.#{name} \"#{xpath}\"#{comma}"}
          end
        end
      end
      @result += @full_definition
    end    
    
    def self.export_footer(output_file, wrapper_name, extractor_result_file_name)
      if extractor_result_file_name
        @result += "\n\n#{wrapper_name}.to_xml.write(open('result_of_exported_extractor.xml', 'w'), 1)"
      else
        @result += "\n\n#{wrapper_name}.to_xml.write($stdout, 1)"
      end
    end

    
    def self.create_name_to_xpath_map(pattern)
      @name_to_xpath_map[pattern.name] = []
      pattern.filters.each do |filter|
        @name_to_xpath_map[pattern.name] << filter.xpath if pattern.filters[0].xpath != nil
      end
      pattern.children.each {|child| create_name_to_xpath_map child}      
    end
    
    def self.replace_example_with_xpath(name, xpaths, left_delimiter, right_delimiter=left_delimiter)
      return if name=='root'
      parens = @full_definition.scan(/P.#{name}\s*\((.+?)\)/)
      if parens.empty?
        full_line = @full_definition.scan(/P.#{name}\W(.+)$/)[0][0]
      else
        full_line = parens[0][0]        
      end
      examples = full_line.split(",")      
      examples.reject! {|exa| exa.strip!;  exa[0..0] != %q{"} && exa[0..0] != %q{'} }
      all_xpaths = ""
      examples.each do |e|
        index = examples.index(e)
        xpath = xpaths[index]
        return if xpath == nil
        all_xpaths += ", " if index > 0
        all_xpaths += '"' + xpath + '"'
      end
      replacing_xpath = full_line.include?('{') ? "P.#{name}('#{all_xpaths}')" :
                                                  "P.#{name} #{all_xpaths}"
      optional_paren_escaped = parens.empty? ? '' : '\('
      optional_paren = parens.empty? ? '' : '('
      @full_definition.sub!(/P\.#{name}\s*#{optional_paren_escaped}#{left_delimiter}(.*)#{right_delimiter}/) do
        @name_to_xpath_map.delete("#{name}")
        optional_paren + replacing_xpath
      end                                                  
    end
    
  end
end
