require 'rubygems'
require 'hpricot'

module Scrubyt
  ##
  #=<tt>Various XPath utility functions</tt>
  class XPathUtils
    #When looking up examples, do NOT recurse into these tags since they won't contain any usable info
    NON_CONTENT_TAGS = ['form','option', 'input', 'script', 'noscript']
    ENTITIES = {
        'quot'      => '"',
        'apos'      => "'",
        'amp'       => '&',
        'lt'        => '<',
        'gt'        => '>',
        'nbsp'      => ' '}
    
    #From the example text defined by the user, find the lowest possible node with the text 'text'.
    #The text can be also a mixed content text, e.g.
    #
    # <a>Bon <b>nuit</b>, monsieur!</a>
    #
    #In this case, <a>'s text is considered to be "Bon nuit, monsieur"
    def self.find_node_from_text(doc, text, next_link)
      @node = nil
      @found = false
      #digg next page hack
      text.gsub!('»', '&#187;')
      self.traverse_for_full_text(doc,text)
      self.lowest_possible_node_with_text(@node, text) if @node != nil
      if (@found == false)        
        #Fallback to per node text lookup
        self.traverse_for_node_text(doc,text)
        if (@found == false)
          return nil if next_link
          puts "!" * 65
          puts "!!!!!! FATAL: Node for example #{text} Not found! !!!!!!" 
          puts "!!!!!! Please make sure you specified the example properly !!!!!!"
          puts "!" * 65
          exit
        end
      end
      @node
    end
    
    #Full text of the node; this is equivalent to Hpricot's inner_text 
    #(? be sure to check). Will be
    #replaced if Hpricot 0.5 will be released
    def self.full_text(node)
      result = ""
      node.traverse_text { |t| result += t.to_s }
      result
    end
     
    #Find the LCA (Lowest Common Ancestor) of two nodes   
    def self.lowest_common_ancestor(node1, node2)
      path1 = traverse_up(node1)
      path2 = traverse_up(node2)
      return node1.parent if path1 == path2

      closure = nil
      while (!path1.empty? && !path2.empty?)
	    closure = path1.pop
	    return closure.parent if (closure != path2.pop)
      end
      path1.size > path2.size ? path1.last.parent : path2.last.parent
    end
           
    ##
    #Generate XPath for the given node
    #
    #*parameters*
    #
    #_node_ - The node we are looking the XPath for
    #
    #_stopnode_ - The Xpath generation is stopped and the XPath that
    #was generated so far is returned if this node is reached. 
    #
    #_write_indices_ - whether the index inside the parent shuold be
    #added, as in html[1]/body[1]/table[2]/tr[1]/td[8]
    def self.generate_XPath(node, stopnode=nil, write_indices=false)
      path = []
      indices = []
      found = false
      while node.class != Hpricot::Doc do
        if node == stopnode
          found = true
          break
        end
        path.push node.name
        indices.push find_index(node) if write_indices
        node = node.parent
      end
      #This condition ensures that if there is a stopnode, and we did not found it along the way,
      #we return nil (since the stopnode is not contained in the path at all)
      return nil if stopnode != nil && !found
      result = ""
      if write_indices
        path.reverse.zip(indices.reverse).each { |node,index| result += "#{node}[#{index}]/" }
      else
        path.reverse.each{ |node| result += "#{node}/" }
      end 
      "/" + result.chop
    end
    
    #Generate an XPath of the node with indices, relatively to the given 
    #relative_root.
    #
    #For example if the elem's absolute XPath is /a/b/c, 
    #and the relative root's Xpath is a/b, the result of the function will
    #be /c.
    def self.generate_relative_XPath( elem,relative_root )
      return nil if (elem == relative_root)
      generate_XPath(elem, relative_root, true)
    end
    
    #Generate a generalized XPath (i.e. without indices) of the node, 
    #relatively to the given relative_root.
    #
    #For example if the elem's absolute XPath is /a[1]/b[3]/c[5], 
    #and the relative root's Xpath is a[1]/b[3], the result of the function will
    #be /c.    
    def self.generate_generalized_relative_XPath( elem,relative_root )
      return nil if (elem == relative_root)
      generate_XPath(elem, relative_root, false)
    end
    
    #Find an image based on the src of the example
    #
    #*parameters*
    #
    #_doc_ - The containing document
    #
    #_example_ - The value of the src attribute of the img tag
    #This is convenient, since if the users rigth-clicks an image and
    #copies image location, this string will be copied to the clipboard
    #and thus can be easily pasted as an examle
    #
    #_index_ - there might be more images with the same src on the page -
    #most typically the user will need the 0th - but if this is not the 
    #case, there is the possibility to override this
    def self.find_image(doc, example, index=0)
      (doc/"//img[@src='#{example}']")[index]
    end

    ##
    #Used to find the parent of a node with the given name - for example
    #find the <form> node which is the parent of the <input> node    
    def self.traverse_up_until_name(node, name)
      while node.class != Hpricot::Doc do
        break if node.name == name
        node = node.parent
      end
      node
    end
    
    
private
    #Find the index of the child inside the parent
    #For example:
    #
    #         tr
    #      /  |   \
    #    td   td   td
    #    0    1    2    
    #
    #The last row contains the indices of the td's from the
    #tow above.
    #
    #Note that in classic XPath, the indices start with 1 (rather
    #than 0).
    def self.find_index(node)
     c = 0
     node.parent.children.each do |child|
       if child.class == Hpricot::Elem
         c += 1 if (child.name == node.name)
         break if (node == child)    
       end
     end
     c
    end

    def self.traverse_up(node, stopnode=nil)
      path = []
      while node.class != Hpricot::Doc do
        break if node == stopnode
        path.push node
        node = node.parent
      end
    path
    end

    def self.traverse_for_node_text(node, text)
      return if @found
      if (node.instance_of? Hpricot::Elem)
        node.traverse_text do |t| 
          if (t.to_s == text)
            @found = true 
            @node = t.parent
          end
       end
      end      
      node.children.each do |child|
        if child.instance_of? Hpricot::Elem
          traverse_for_node_text(child, text) unless NON_CONTENT_TAGS.include? child.name
        end
      end      
    end

    def self.traverse_for_full_text(node, text)
      return if @found
      if (node.instance_of? Hpricot::Elem)  
        ft = unescape_entities(full_text(node)).strip
        #puts ">>#{ft}<<, text:>>#{text}<<" if ((ft.size < 20) && ft =~ /Open Source/)        
        if (ft == text)
          @found = true 
          @node = node
        end
      end
      node.children.each do |child|
        if child.instance_of? Hpricot::Elem
          traverse_for_full_text(child, text) unless NON_CONTENT_TAGS.include? child.name
        end
      end
    end
    
    def self.unescape_entities(text)
        ENTITIES.each {|e,s| text.gsub!(/\&#{e};/) {"#{s}"} }
        text
    end

    def self.lowest_possible_node_with_text(node, text)
      return if node.instance_of? Hpricot::Text
      @node = node if full_text(node) == text
      node.children.each  { |child| lowest_possible_node_with_text(child, text) }
    end #End of method lowest_possible_node_with_text    
  end #End of class XPathUtils
end #End of module Scrubyt
